#!/usr/bin/env python

"""Console script for ibdpainting."""

import argparse
from ibdpainting import __version__

def main():
    parser = argparse.ArgumentParser(description='ibdpainting')

    parser.add_argument('-i', '--input',
        help='Path to an HDF5 file containing genotype data for one or more samples to check. This should be the output of allel.vcf_to_hdf5().'
        )
    parser.add_argument('-n', '--sample_name',
        help ='Sample name for the individual to check. This must be present in the samples in the input file.'
    )
    parser.add_argument('-r', '--reference',
        help="Path to an HDF5 file containing genotype data for a panel of reference individuals to compare the input indivual against. This should be the output of allel.vcf_to_hdf5()."
    )
    parser.add_argument('-w', '--window_size',
        type=int, default=500000,
        help="Integer window size in base pairs. Defaults to 500000 bp."
    )
    parser.add_argument('--expected_match',
        help="Optional list of sample names in the reference panel that are expected to be ancestors of the test individual.",
        nargs = "+", required=False, default=[]
    )
    parser.add_argument('--outdir',
        help="Directory to save the output."
    )
    parser.add_argument('--keep_ibd_table', 
        help="If set, write an intermediate text file giving genetic distance between the crossed individual and each candidate at each window in the genome. Defaults to False, because these can be quite large.",
        default=False,
        action=argparse.BooleanOptionalAction
    )
    parser.add_argument('--max_to_plot', 
        help="Optional number of the best matching candidates to plot so that the HTML files do not get too large and complicated. Ignored if this is more than the number of samples. Defaults to 20.",
        type=int, default = 10
    )
    parser.add_argument('--interactive',
        help="If set, save the output plot as an interactive HTML plot including information on candidates within the plot.",
        default=True,
        action=argparse.BooleanOptionalAction
        )
    parser.add_argument('--height',
        help="Height in pixels of the output PNG file. Defaults to 675.",
        default=675
        )
    parser.add_argument('--width',
        help="Height in pixels of the output PNG file. Defaults to 900.",
        default=900)
    parser.add_argument('--version',
        action='version',
        version=f'%(prog)s {__version__}'
        )
    
    args = parser.parse_args()

    # Only import and run analysis if actually needed
    if len(vars(args)) > 1:  # If there are actual arguments besides --help
        run_analysis(args)
    else:
        parser.print_help()

def run_analysis(args):
    import os
    
    from ibdpainting.ibd_table import ibd_table
    from ibdpainting.ibd_scores import ibd_scores
    from ibdpainting.plot_ibd_table import plot_ibd_table
    
    if not os.path.isdir(args.outdir):
        os.makedirs(args.outdir)

    # Data frame of IBD at all positions across the genome, and the plot of this
    itable = ibd_table(
        input = args.input,
        reference=args.reference,
        sample_name=args.sample_name,
        expected_match=args.expected_match,
        window_size=args.window_size
        )
    
    print("Calculating the ibd_scores table.")
    scores = ibd_scores(itable)
    scores.to_csv( args.outdir + "/" + args.sample_name + "_ibd_scores.csv", index=False)
       
    print("Creating the plot.")
    fig = plot_ibd_table(
        itable,
        args.sample_name,
        args.expected_match,
        args.max_to_plot
        )
    png_out = args.outdir + "/" + args.sample_name + "_plot_ibd.png"
    print(f"Writing to {png_out}")
    fig.write_image(
        png_out,
        height = args.height, width = args.width
        )
    
    if args.interactive:
        html_out = args.outdir + "/" + args.sample_name + "_plot_ibd.html"
        print(f"Writing an interactive plot to {html_out}.")
        fig.write_html(args.outdir + "/" + args.sample_name + "_plot_ibd.html")
                
    if args.keep_ibd_table:
        itable_path = args.outdir + "/" + args.sample_name + "_ibd_table.csv"
        print(f"Writing the full ibd_table to {itable_path}")
        itable.to_csv(itable_path, index=False)
    

if __name__ == '__main__':
    main()



import ibdpainting as ip
import numpy as np

input = 'tests/test_data/panel_to_test.hdf5'
reference = 'tests/test_data/reference_panel.hdf5'
ref_vcf = 'tests/test_data/reference_panel.hdf5'
chr1 = 'tests/test_data/reference_panel_chr1.hdf5'

def test_ibd_table():
    ibd = ip.ibd_table(
        input=ref_vcf,
        reference=reference,
        sample_name='1158',
        window_size=1000
    )
    # Check the dataframe is the right shape
    assert ibd.shape == (200, 5)
    # Check the column names are as expected
    assert ibd.keys().to_list() == ['window', '1158', '6024', '6184', '8249']
    # Check that the column for the true parent is all zeroes or -9
    assert all(
        (ibd['1158'] == 1) | (np.isnan(ibd['1158']))
    )
    # Check that a non-parent are not all -9.
    assert any(ibd['8249'] != 0)


def test_ibdtable_returns_expected_F1():
    ibd = ip.ibd_table(
        input=input,
        reference=reference,
        sample_name='F2.05.015',
        expected_match=['6024', '8249'],
        window_size=1000
    )

    # Check the column names are as expected
    assert ibd.keys().to_list() == ['window', '1158', '6024', '6184', '8249', 'Expected_F1']



import numpy as np
import pandas as pd
import warnings
import itertools

def ibd_scores(ibd_table, rank_threshold:int=100):
    """
    Genetic distance between offspring and parents for pairs of plausible 
    candidate parents.

    This starts with the premise that crossed individuals can be thought of as 
    a mosaic of the two parents, such that in any single window they ought to be
    identical to one, but not the other parent.
    
    Beginning with a table of genetic distances between a test individual and
    each genotype in a reference panel, this function compares all plausible 
    pairs of candidates and pulls out the *maximum* genetic similarity to the 
    offspring in each window. It then takes the average over windows, excluding
    windows with NA values.
    
    Because there can be a very large number of pairs to consider, candidates 
    are first filtered to retain the N candidates with the lowest mean genetic
    distance from the offspring. N is controlled by the argument
    `rank_threshold`.

    Parameters
    ==========
    ibd_table: pd.DataFrame
        DataFrame with a row for each window in the genome and a column for each 
        sample in the reference panel. Elements show genetic distance between the 
        test individual and each reference individual in a single window.
        This is generated by ibd_table().
    rank_threshold: int, optional
        Number of most-plausible to consider as pairs. Defaults to 20. Setting 
        this number to something large will make the output very large!

    Returns
    =======
    A DataFrame with a row for each pair of candidate parents, giving candidate
    IDs and the average mimimum genetic distance from either parent to the
    offspring, averaged over windows. Values closer to zero indicate that the 
    pair is more likely to be a match.
    """
    # Coerce missing data to NaN for correct column means.
    ibd_table = ibd_table.replace(-9,np.nan)
    
    # Get column-mean IBD for each candidate, allowing for missing data
    # Note the index from 1 onwards to skip the window column
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ibd_scores_for_each_candidate = np.array(
            [ np.nanmean(ibd_table[col]) for col in ibd_table.keys()[1:] ]
        )
    # Get the indices of the top N values in ibd_scores_for_each_candidate
    # where N is rank_threshold
    if rank_threshold < ibd_table.shape[1]:
        score_ix = np.argpartition(ibd_scores_for_each_candidate, -rank_threshold)[-rank_threshold:]
    else:
        score_ix = np.arange(len(ibd_scores_for_each_candidate))
    
       
    # Reference sample names, excluding the column for window
    all_names = ibd_table.keys()[1:] # All names in the reference panel
    ref_names = all_names[score_ix] # Top candidates
    # For calculating scores for pairs we don't want combinations with 'Expected_F1'
    pair_names = list(ref_names)
    if 'Expected_F1' in pair_names:
        pair_names.remove("Expected_F1")
    
    # Dataframe giving the scores if parents are selfed.
    score_table = pd.DataFrame({
        'parent1'    : ref_names,
        'parent2'    : ref_names, # Under selfing Parent1 and Parent2 are the same. 
        'similarity' : ibd_scores_for_each_candidate[score_ix]
    })
    
    # Get scores for all pairs of top parents.
    # This is the maxmimum similarity score for either parent in a pair in each
    # window.
    # No pairs should contain "Expected_F1" as a candidate.
    if rank_threshold > 1:
        scores_for_pairs = []
        for i,j in itertools.combinations(pair_names,2):
            # The maximum similarity at each window between the test individual and 
            # *either* reference genotype in this pair
            max_similarity_pair = ibd_table[[i,j]].max(axis=1, skipna=False)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # List containing the ID of candidates 1 and 2 and the average of
                # the best similarities at each window.
                scores_for_pairs.append(
                    [ i,j, np.nanmean(max_similarity_pair) ]
                    )
        scores_for_pairs = pd.DataFrame(scores_for_pairs, columns = ['parent1', 'parent2', 'similarity'])
    
        # A single dataframe combining scores for selfed and pairs.
        score_table = pd.concat([score_table, scores_for_pairs])
        
    # Sort so the best matches are at the top
    score_table = score_table.sort_values('similarity', ascending=False)
    
    return score_table


import subprocess


input = 'tests/test_data/panel_to_test.hdf5'
reference = 'tests/test_data/reference_panel.hdf5'
ref_vcf = 'tests/test_data/reference_panel.hdf5'
chr1 = 'tests/test_data/reference_panel_chr1.hdf5'

def test_cli_version():
    result = subprocess.run(
        ['ibdpainting', '--version'],
        capture_output=True,
        text=True
    )
    assert result.returncode == 0
    assert "ibdpainting" in result.stdout

"""
ibdpainting \
    --input tests/test_data/reference_panel.hdf5 \
    --reference tests/test_data/reference_panel.hdf5 \
    --sample_name 1158 \
    --window_size 1000 \
    --outdir tests/test_output \
    --expected_match 1158 \
    --no-interactive
"""


import pandas as pd
import numpy as np

import plotly.express as px

def plot_ibd_table(ibd_table:pd.DataFrame, sample_name:str, expected_match:list=[], max_to_plot=10):
    """
    Plot allele sharing across the genome.

    Create a interactive line graph showing genetic distance from a test
    individual to each sample in a panel of reference individuals.

    Parameters
    ==========
    ibd_table: pd.DataFrame
        DataFrame with a row for each window in the genome and a column for each 
        sample in the reference panel. Elements show genetic distance between the 
        test individual and each reference individual in a single window.
        This is generated by ibd_table().
    sample_name: str
        Sample name for the individual to check.
        This must be present in the samples in the input VCF.
    expected_match: list
        List of sample names in the reference panel that are expected to be
        ancestors of the test individual.

    Returns
    =======
    Plotly figure object with subplots for each chromosome, showing window
    position along the x-axis and genetic distance from the test individual to
    each reference sample on the y-axis. Line colour indicates whether a sample
    is an expected parent or not. Rolling over the lines shows which sample is
    which.
    """
    # Add the expected F1 to the list of expected matches so that it is plotted in colour
    if len(expected_match) == 2:
        expected_match = expected_match + ['Expected_F1']

    # Identify the candidate names *not* among the top `max_to_plot` columns and remove
    # If `max_to_plot` is less than the number of candidates.
    if max_to_plot < ibd_table.shape[1]-1:
        # Get column-mean IBD for each candidate, allowing for missing data
        ibd_scores_for_each_candidate = np.array(
            [ np.nanmean(ibd_table[col]) for col in ibd_table.keys()[1:] ]
        )
        # Identify the candidate names *not* among the top `max_to_plot` columns
        ix = np.argpartition(-ibd_scores_for_each_candidate, max_to_plot)[max_to_plot:] # index positions
        columns_to_drop = ibd_table.keys()[ix+1].to_list() # candidate names to be removed
        # Make sure the expected parents are not among the proscribed
        [ columns_to_drop.remove(x) for x in expected_match if x in columns_to_drop ]        
        ibd_table = ibd_table.drop(columns=columns_to_drop) # drop the candidates

    # Make the table long
    ibd_table = ibd_table.melt(id_vars=['window'], var_name='candidate', value_name='similarity')
    # Column indicating which candidates should be plotted a different colour.
    ibd_table['colour'] = np.where(
        ibd_table['candidate'].isin(expected_match), ibd_table['candidate'], "Other"
        )
    # Unique list of labels for the legend, sorted to plot "Other" first, then
    # then the expected matches (including heterozygosity, if using).
    unique_legend_labels = ['Other'] + expected_match

    # Split the 'window' column up into separate columns for chromosome, start and stop positions
    ibd_table[['chr', 'window']] = ibd_table['window'].str.split(":", expand=True)
    ibd_table[['start', 'stop']] = ibd_table['window'].str.split("-", expand=True)
    # start and stop positions should be integers for sensible plotting.
    ibd_table['start'] = ibd_table['start'].astype(int)
    ibd_table['stop'] = ibd_table['stop'].astype(int)
    ibd_table['midpoint'] = (ibd_table['start'] + ibd_table['stop']) / 2
        
    # Set the palette for the lines in the plot.
    # This sets 'Others', then uses the default palette for the first N colours
    custom_colors = ['LightGrey'] + px.colors.qualitative.Set1[0:len(unique_legend_labels)]

    # Make the plot
    fig = px.line(
        ibd_table,
        x="midpoint", y="similarity", color="colour", line_group="candidate",
        title=sample_name,
        labels={
            'midpoint' : 'Position (bp)',
            'distance' : 'Similarity'        
        },
        hover_data=['candidate'],
        template="plotly_white",
        color_discrete_sequence=custom_colors,
        category_orders={'colour': unique_legend_labels},
        facet_row = "chr"
        )
    fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[-1])) # Remove "chr=" from facet labels
    fig.update_layout(legend_title_text='Candidate')
    fig.update_traces(mode="markers+lines")

    return fig



import h5py
import numpy as np

def find_matching_markers(input: str, reference: str, sample_name: str) -> dict:
    """
    Prepare HDF5 files for further processing.

    Imports HDF5 files for the test and reference panels and run initial checks.
    This then compares the lists of SNPs in each panel and identifies those that
    are shared.

    Parameters
    ----------
    input : str
        Path to a an HDF5 file containing genotype data for one or more samples to check
    reference : str
        Path to a HDF5 file containing genotype data for a panel of reference individuals
        to compare the input indivual against.
    sample_name : str
        Sample name for the individual to check. This must be present in the samples
        in the input file.
    
    Returns
    -------
    A dictionary listing:
    samples : np.array
        A vector of strings giving the test individual and all reference
        individuals.
    chr : np.array
        Array of chromosome labels for shared SNPs
    pos : np.array
        Array of positions labels for shared SNPs
    """
    # Read in the data files
    input_hdf5 = h5py.File(input, mode='r')
    ref_hdf5  = h5py.File(reference, mode="r")

    # The HDF5 files contain byte strings
    # Convert them back to regular strings
    input_str_data = {
        'samples' : [ x.decode('utf-8') for x in input_hdf5['samples'][:] ],
        'chr'     : [ x.decode('utf-8') for x in input_hdf5['variants/CHROM'][:]]
    }
    ref_str_data = {
        'samples' : [ x.decode('utf-8') for x in ref_hdf5['samples'][:] ],
        'chr'     : [ x.decode('utf-8') for x in ref_hdf5['variants/CHROM'][:] ]
    }
    print(f"The reference panel contains {len(ref_str_data['samples'])} samples and {len(ref_str_data['chr'])} loci.")
    print(f"The test panel contains {len(input_str_data['samples'])} samples and {len(input_str_data['chr'])} loci.")

    if sample_name not in input_str_data['samples']:
        raise ValueError("The sample name is not in the list of samples in the input file.")
    else: 
        print(f"Sample {sample_name} identified correctly in the list of samples in the input file.")
        # Find the position of the individual to test
        sample_ix = np.where(
            [ sample_name == x for x in input_str_data['samples'] ]
            )[0][0]
        # Join vectors of sample names, with the test individual first
        new_samples = np.append(
            input_str_data['samples'][sample_ix], ref_str_data['samples']
            )

    # Check that contig labels match
    chr_labels = {
        'input' : np.unique(input_str_data['chr']),
        'ref'   : np.unique(ref_str_data['chr'])
    }
    if len(chr_labels['input']) != len(chr_labels['ref']):
        raise ValueError(
            "The number of unique contig labels do not match: the input an HDF5 has {}, but the reference panel has {}.".
            format( chr_labels['input'], chr_labels['ref'] )
        )
    elif any( chr_labels['input'] != chr_labels['ref'] ):
        raise ValueError(
            "Contig labels do not match between the input and reference files."
        )
    else:
        print("Contig labels seem to match between the input and reference panels.")

    # Make sure we only compare SNPs that are found in both datasets.
    # Concatenate chromosome labels and SNP positions
    snp_names = {
        'input' : [ str(chr) + ":" + str(pos) for chr,pos in zip(input_str_data['chr'], input_hdf5['variants/POS'][:]) ],
        'ref'   : [ str(chr) + ":" + str(pos) for chr,pos in zip(ref_str_data['chr'], ref_hdf5['variants/POS'][:]) ]
    }
    # Find the SNP position names that are common to both datasets
    matching_SNPs_in_both_files = set(set(snp_names['input']) & set(snp_names['ref']))
    which_SNPs_to_keep = {
        "input" : [ x in matching_SNPs_in_both_files for x in snp_names['input'] ],
        "ref"   : [ x in matching_SNPs_in_both_files for x in snp_names['ref'] ]
    }
    print(f"{len(matching_SNPs_in_both_files)} markers are found in both the input and reference panels.")

    output = {
        'sample_ix' : sample_ix,
        'input' : which_SNPs_to_keep['input'],
        'ref' : which_SNPs_to_keep['ref']
        }
    
    input_hdf5.close()
    ref_hdf5.close()

    return output



import ibdpainting as ip
import numpy as np


input = 'tests/test_data/panel_to_test.hdf5'
reference = 'tests/test_data/reference_panel.hdf5'
ref_vcf = 'tests/test_data/reference_panel.hdf5'
chr1 = 'tests/test_data/reference_panel_chr1.hdf5'

ibd = ip.ibd_table(
    input=input,
    reference=reference,
    sample_name='S2.06.002',
    expected_match=['6024'],
    window_size=10000
)


def test_ibd_scores_works():
    """Test basic functionality of ibd_scores
    """
    
    scores = ip.ibd_scores(ibd)
    assert scores.shape == (10,3)
    
def test_setting_rankthreshold_works():
    """Check that ibd_scores works when you mess with rank_threshold,
    """
    assert ip.ibd_scores(ibd, rank_threshold=1).shape == (1,3)
    assert ip.ibd_scores(ibd, rank_threshold=3).shape == (6,3)
    assert ip.ibd_scores(ibd, rank_threshold=4).shape == (10,3)
    assert ip.ibd_scores(ibd, rank_threshold=100).shape == (10,3)

def test_ibd_scores_propogates_NA():
    """
    Test that missing values inhereited from geneticDistance as -9 are
    propogated as NaN
    """
    ibd = ip.ibd_table(
        input=input,
        reference=reference,
        sample_name='S2.06.002',
        window_size=1000
    )
    # Set values for 1158 as missing
    ibd['1158'] = -9
    ibd = ibd.replace(-9, np.nan)
    
    scores = ip.ibd_scores(ibd)
    
    assert all(
        scores.loc[scores['parent1'] == "1158"]['similarity'].isna()
    )



import h5py
import pandas as pd
import numpy as np
import numpy.ma as ma
from tqdm import tqdm


from ibdpainting.find_matching_markers import find_matching_markers


def pairwise_distance(geno, ref_sample_names, expected_match:list[str]=[]):
    """
    Calculate pairwise genetic distance between an input individual and all 
    reference individuals.

    The input individual is always the first in the list of samples. Genetic
    distance is the number of allelic differences at each locus between each
    pair, summed over all loci. The calculation is done using masked arrays to
    account for missing data.
    
    If a list of two expected parents is given as `expected_match`, this adds an
    additional reference individual as the expected genotype of the F1 between
    those parents.
    
    This actually returns one minus the genetic distances, so that perfect
    matches have a score of one.
    
    Returns
    =======
    Vector of similarities between the sample and every reference genotype.
    The final element is the similarity with the expected genotype of an F1 
    between the expected parents, if used.
    The similarity is one minus the mean of distances across loci, excluding
    cases with missing data.

    """
    # NxM array of genotypes, with missing values masked.
    masked_geno = ma.masked_array(geno, geno < 0)
    # NxM array of positions where either allele is missing
    mask_any = masked_geno.mask.any(axis=2)

    # Nx1 Vector of diploid genotypes for the sample
    sample_geno      = masked_geno[:,0].sum(1)
    sample_geno.mask = mask_any[:,0] # Mask SNP positions with any missing data
    # Nx(M-1) Array of diploid genotypes for the reference panel
    ref_geno      = masked_geno[:,1:].sum(2)
    ref_geno.mask = mask_any[:,1:]
    
    # If there are exactly two expected parents, add an additional reference
    # genotype, corresponding to the expected genotype of the F1 between those
    # lines.
    if len(expected_match) == 2:
        # Indices of the expected parents
        # The +1 is there because masked_geno has an extra column in position 0.
        expected_ix = [ref_sample_names.index(item)+1 for item in expected_match]
        # Diploid genotypes of the expected parents
        exp_diploid      = masked_geno[:,expected_ix].sum(2)
        exp_diploid.mask = masked_geno[:,expected_ix].mask.any(axis=2)
        # Calculate the expected genotype of an F1
        exp_F1 = exp_diploid.sum(axis=1)
        exp_F1.mask = np.ma.getmaskarray(exp_diploid).any(axis=1)
        exp_F1 = exp_F1 / 2
        # When one of the expected parents is heterozygous we can't know what the 
        # expected genotype will be. This means the expected genotype will be a float
        # not equal to 0, 1 or 2. Set these calls to NA.
        exp_F1.mask = ma.getmaskarray(exp_F1) | ~np.isin(exp_F1.data, np.array([0,1,2]))
        # Ensure data type matches the reference genotypes.
        exp_F1 = exp_F1.astype(ref_geno.dtype)
        # Concatenate expected F1 genotype to the reference panel.
        ref_geno = ma.hstack([ref_geno, exp_F1[:,np.newaxis]])
    
    # Array of genetic distances from the sample to each candidate, including the
    # pseudoheterozygote
    difference_array = abs(sample_geno[:,np.newaxis] - ref_geno) / 2
    # A vector of average distances for each reference candidate
    difference_vector = difference_array.mean(axis=0)
    
    # Return the result as a vector of *similarities*
    return ma.filled(1-difference_vector, np.nan)


def genotype_calls_by_window(input_hdf5, ref_hdf5, chr, start, stop, matching_markers):
    """
    Extract a table of genotype calls for a single window.
    
    This indexes genotype calls from and entire HDF5 file using boolean arrays
    indexing chromosome, positions inside a window, and whether the SNPs are
    shared between the test and reference panels.
        
    Returns
    -------
    An NxMx2-dimensional array where
    - N is the number of SNPs in a single window
    - M is columns for the test individual followed by all reference
    - 2 indexes the two chromosomes.
    
    Missing data are coded as negative integers.
    """
    # Create a matrix of genotype data for the test and reference individuals.
    # Arrays of Booleans matching the current chromosome
    chr_ix_test = (input_hdf5['variants/CHROM'][:] == chr)
    chr_ix_ref  = (ref_hdf5['variants/CHROM'][:] == chr)
    # Arrays matching the current window.
    window_ix_test = (input_hdf5['variants/POS'][:] >= start) & (input_hdf5['variants/POS'][:] < stop)
    window_ix_ref  = (ref_hdf5['variants/POS'][:] >= start)   & (ref_hdf5['variants/POS'][:] < stop)
    # Match chromosome and window to shared SNPs
    input_ix_test = chr_ix_test & window_ix_test & matching_markers['input']
    input_ix_ref  = chr_ix_ref  & window_ix_ref  & matching_markers['ref']
    # Take shared SNPs in this window, for the test individual only.
    input_genotypes = input_hdf5['calldata/GT'][input_ix_test, matching_markers['sample_ix']]
    # Shared SNPs in this window for the whole test panel.
    ref_genotypes = ref_hdf5['calldata/GT'][input_ix_ref]
        
    if(ref_genotypes.shape[0] != input_genotypes.shape[0]):
        raise ValueError(f"The test panel at this window has {input_genotypes.shape[0]} loci , but the reference panel has {ref_genotypes.shape[0]}.")
    else:
        # Empty matrix to store genotype data.
        # The first row should contain genotypes for the test individual
        # Subsequent rows should contain genotypes for all reference indivdiduals.
        new_geno = np.empty((ref_genotypes.shape[0], ref_genotypes.shape[1] +1, 2), dtype=input_hdf5['calldata/GT'].dtype)
        new_geno[:, 0]  = input_genotypes
        new_geno[:, 1:] = ref_genotypes
    
    return new_geno


def ibd_table(input:str, reference:str, sample_name:str, expected_match:list[str]=[], window_size:int=500000):
    """
    Compare allele sharing across the genome.

    Calculate genetic distance between a test individual and a panel of
    reference genomes.

    Parameters
    ==========
    input: str
        Path to an HDF5 file containing genotype data for one or more samples to
        test
    reference: str
        Path to an HDF5 file containing genotype data for a panel of reference
        individuals to compare the test individual against.
    sample_name: str
        Sample name for the individual to check.
        This must be present in the samples in the input HDF5 file.
    expected_match: List of str
        List of names (as strings) of the expected parents. If there are exactly
        two expected parents, the expected genotype of their F1 will be included
        as an additional reference panel.
    window_size: int
        Window size in base pairs.

    Returns
    =======
    DataFrame with a row for each window in the genome and a column for each 
    sample in the reference panel. Elements show genetic distance between the 
    test individual and each reference individual in a single window.
    """

    # Read in the data files
    input_hdf5 = h5py.File(input, mode='r')
    ref_hdf5  = h5py.File(reference, mode="r")
    
    # Identify markers that are shared between the test and reference panels
    matching_markers = find_matching_markers(
        input, reference, sample_name
    )
    
    # Vector of sample names for the reference panel
    ref_sample_names = [ x.decode('utf-8') for x in ref_hdf5['samples'][:]]
    # Header for the output dataframe
    genetic_distances_header = ['window','chr', 'start', 'stop'] + ref_sample_names
    # If there are two expected parents, add an extra header for the expected F1
    if len(expected_match) == 2:
        genetic_distances_header = genetic_distances_header + ['Expected_F1']

    # Empty dict to store genetic distances (vectors)
    # Indexed by window_name
    genetic_distances = []
    for chr in np.unique(input_hdf5['variants/CHROM'][:]):
        print(f"Calculating genetic distances across {chr.decode('utf-8')}.")
        
        chr_size = ref_hdf5['variants/POS'][ref_hdf5['variants/CHROM'][:] == chr].max()
        start_positions = np.arange(0, chr_size, window_size)
        
        for start in tqdm(start_positions):
            stop  = start + window_size
            window = chr.decode('utf-8') + ":" + str(start) + "-" + str(stop)
            # Matrix of genotype calls
            geno = genotype_calls_by_window(input_hdf5, ref_hdf5, chr, start, stop,  matching_markers)
            # Vector of genetic distances from the test individual to every reference.
            distances = pairwise_distance(geno, ref_sample_names, expected_match)
            
            # Dataframe with a single row, giving window position, distances
            genetic_distances.append(pd.DataFrame(
                [[window, chr.decode('utf-8'), start, stop] + distances.tolist() ],
                columns = genetic_distances_header
                )
            )
        
    # Create a single dataframe
    genetic_distances = pd.concat(genetic_distances)
    # Sort by chromosome and window position
    # This isn' really necessary here, but would be if I added multiprocessing later.
    genetic_distances = genetic_distances.sort_values(by=['chr', 'start'])
    
    # Column gymnastics to ensure we have columns 'window', then one col for each
    # reference individual.
    genetic_distances = genetic_distances.drop(columns=['chr', 'start', 'stop'])
    
    input_hdf5.close()
    ref_hdf5.close()
    
    return genetic_distances

# ibd_tablex = ibd_table(input, reference, sample_name, window_size)



"""Top-level package for ibdpainting."""

__author__ = """Tom Ellis"""
__email__ = 'thomas.ellis@gmi.oeaw.ac.at'
__version__ = '0.8.4'

import sys

# Only import everything if we're not running from command line
if not sys.argv[0].endswith('ibdpainting'):
    from ibdpainting.find_matching_markers import find_matching_markers
    from ibdpainting.ibd_table import ibd_table
    from ibdpainting.ibd_scores import ibd_scores
    from ibdpainting.plot_ibd_table import plot_ibd_table

    __all__ = [
        'load_genotype_data',
        'ibd_table',
        'ibd_scores',
        'plot_ibd_table'
    ]
